from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
from collections import namedtuple
from enum import Enum
import itertools
import json
from six import string_types
from six.moves.urllib import parse
from sqlalchemy import create_engine
import pandas as pd
import requests

from superset.db_engines.ssas.exceptions import Error, NotSupportedError, ProgrammingError
from olap.xmla.xmla import XMLAProvider

from requests.auth import HTTPBasicAuth
from requests_ntlm import HttpNtlmAuth
import sqlparse

logger = logging.getLogger(__name__)


class Type(Enum):
    STRING = 1
    NUMBER = 2
    BOOLEAN = 3


def connect(host='localhost', port=80, path='/OLAP/msmdpump.dll', scheme='http', username=None, password=None,
            trusted_connection=False):
    """
    Constructor for creating a connection to the database.

        >>> conn = connect('localhost', 8082)
        >>> curs = conn.cursor()

    """
    return Connection(host, port, path, scheme, username=username, password=password,
                      trusted_connection=trusted_connection)


def check_closed(f):
    """Decorator that checks if connection/cursor is closed."""

    def g(self, *args, **kwargs):
        if self.closed:
            raise Error(f'{self.__class__.__name__} already closed')
        return f(self, *args, **kwargs)

    return g


def check_result(f):
    """Decorator that checks if the cursor has results from `execute`."""

    def g(self, *args, **kwargs):
        if self._results is None:
            raise Error('Called before `execute`')
        return f(self, *args, **kwargs)

    return g


MAX_COLS = 254  # python 3.6 namedtuple constraint


def get_description_from_row(row, res):
    """
    Return description from a single row.

    """
    ret = []
    axis_tuple = res.getAxisTuple(0)
    ret.append(
        (
            "label",  # name
            Type.STRING,  # type_code
            None,  # [display_size]
            None,  # [internal_size]
            None,  # [precision]
            None,  # [scale]
            True,  # [null_ok]
        )
    )
    if len(row) and isinstance(row[0], list):
        row = row[0]

    for i in range(min(len(row), MAX_COLS - 1)):
        c = row[i]
        if len(axis_tuple) > i:
            axis = axis_tuple[i]
            t = get_type(c.get('FmtValue'))
            ret.append(
                (
                    axis.get('Caption'),  # name
                    t,  # type_code
                    None,  # [display_size]
                    None,  # [internal_size]
                    None,  # [precision]
                    None,  # [scale]
                    t == Type.STRING,  # [null_ok]
                )
            )
    return ret


from zeep import ns

nsmap = {
    "soap": ns.SOAP_11,
    "soap-env": ns.SOAP_ENV_11,
    "wsdl": ns.WSDL,
    "xsd": ns.XSD,
    "sql": "urn:schemas-microsoft-com:xml-sql"
}


def get_description_from_rowset(res):
    """
    Return description from a single row.

    """
    ret = []
    for i in range(len(res)):
        c = res[i]

        t = get_type_from_schema(c.get('type'))
        ret.append(
            (
                c.get("name"),  # name
                t,  # type_code
                None,  # [display_size]
                None,  # [internal_size]
                None,  # [precision]
                None,  # [scale]
                t == Type.STRING,  # [null_ok]
            )
        )
    return ret


def get_type(value):
    """Infer type from value."""
    if isinstance(value, string_types) or value is None:
        return Type.STRING
    elif isinstance(value, (int, float)):
        return Type.NUMBER
    elif isinstance(value, bool):
        return Type.BOOLEAN

    raise Error(f'Value of unknown type: {value}')


def get_type_from_schema(t):
    """Infer type from value.
    http://books.xmlschemata.org/relaxng/relax-CHP-19.html
    xsd:anyURI - URI (Uniform Resource Identifier)
    xsd:base64Binary - Binary content coded as "base64"
    xsd:boolean - Boolean (true or false)
    xsd:byte - Signed value of 8 bits
    xsd:date - Gregorian calendar date
    xsd:dateTime - Instant of time (Gregorian calendar)
    xsd:decimal - Decimal numbers
    xsd:double - IEEE 64-bit floating-point
    xsd:duration - Time durations
    xsd:ENTITIES - Whitespace-separated list of unparsed entity references
    xsd:ENTITY - Reference to an unparsed entity
    xsd:float - IEEE 32-bit floating-point
    xsd:gDay - Recurring period of time: monthly day
    xsd:gMonth - Recurring period of time: yearly month
    xsd:gMonthDay - Recurring period of time: yearly day
    xsd:gYear - Period of one year
    xsd:gYearMonth - Period of one month
    xsd:hexBinary - Binary contents coded in hexadecimal
    xsd:ID - Definition of unique identifiers
    xsd:IDREF - Definition of references to unique identifiers
    xsd:IDREFS - Definition of lists of references to unique identifiers
    xsd:int - 32-bit signed integers
    xsd:integer - Signed integers of arbitrary length
    xsd:language - RFC 1766 language codes
    xsd:long - 64-bit signed integers
    xsd:Name - XML 1.O name
    xsd:NCName - Unqualified names
    xsd:negativeInteger - Strictly negative integers of arbitrary length
    xsd:NMTOKEN - XML 1.0 name token (NMTOKEN)
    xsd:NMTOKENS - List of XML 1.0 name tokens (NMTOKEN)
    xsd:nonNegativeInteger - Integers of arbitrary length positive or equal to zero
    xsd:nonPositiveInteger - Integers of arbitrary length negative or equal to zero
    xsd:normalizedString - Whitespace-replaced strings
    xsd:NOTATION - Emulation of the XML 1.0 feature
    xsd:positiveInteger - Strictly positive integers of arbitrary length
    xsd:QName - Namespaces in XML-qualified names
    xsd:short - 32-bit signed integers
    xsd:string - Any string
    xsd:time - Point in time recurring each day
    xsd:token - Whitespace-replaced and collapsed strings
    xsd:unsignedByte - Unsigned value of 8 bits
    xsd:unsignedInt - Unsigned integer of 32 bits
    xsd:unsignedLong - Unsigned integer of 64 bits
    xsd:unsignedShort - Unsigned integer of 16 bits
    """

    if t == "xsd:int" \
            or t == "xsd:long" \
            or t == "xsd:double" \
            or t == "xsd:float " \
            or t == "xsd:decimal " \
            or t == "xsd:unsignedByte" \
            or t == "xsd:unsignedInt" \
            or t == "xsd:unsignedLong" \
            or t == "xsd:unsignedShort" \
            or t == "xsd:short" \
            or t == "xsd:positiveInteger" \
            or t == "xsd:nonPositiveInteger" \
            or t == "xsd:nonNegativeInteger" \
            or t == "xsd:negativeInteger" \
            or t == "xsd:integer" \
            or t == "xsd:gYear" \
            or t == "xsd:gYearMonth" \
            or t == "xsd:gMonthDay" \
            or t == "xsd:gMonth" \
            or t == "xsd:gDay" \
            or t == "xsd:byte" \
            or t == "xsd:gMonthDay" \
            or t == "xsd:gMonthDay" \
            or t == "xsd:gMonthDay" \
            or t == "xsd:gMonthDay" \
            or t == "xsd:gMonthDay ":
        return Type.NUMBER
    elif t == "xsd:boolean":
        return Type.BOOLEAN
    else:
        return Type.STRING


class Connection(object):
    """Connection to a OLAP database."""

    def __init__(
            self,
            host='localhost',
            port=80,
            path='/OLAP/msmdpump.dll',
            scheme='http',
            username=None,
            password=None,
            trusted_connection=False
    ):
        netloc = f'{host}:{port}'
        self.url = parse.urlunparse(
            (scheme, netloc, path, None, None, None))
        self.closed = False
        self.cursors = []
        auth = None
        if trusted_connection and username:
            auth = HttpNtlmAuth(username, password)
        elif username:
            auth = HTTPBasicAuth(username, password)

        provider = XMLAProvider()
        # c = provider.connect(location='http://localhost/OLAP/msmdpump.dll',sslverify=False,  auth = HttpNtlmAuth('EBILET\mfyuce', 'FdfuSadD7cV'))
        self.xmla = provider.connect(location=self.url, sslverify=False, auth=auth)
        # c = provider.connect(location='http://10.134.2.182/OLAP/msmdpump.dll',sslverify=False, auth = HTTPBasicAuth('DwhLinkedUser', 'Besiktas!1903'))
        # a = c.BeginSession()
        # print (c.getDatasources())
        # print (c.getMDSchemaCubes())
        # source = c.getOLAPSource()
        # print (source.getCatalog("TESTforNetAI"))

    @check_closed
    def close(self):
        """Close the connection now."""
        self.closed = True
        for cursor in self.cursors:
            try:
                cursor.close()
            except Error:
                pass  # already closed

    @check_closed
    def commit(self):
        """
        Commit any pending transaction to the database.

        Not supported.
        """
        pass

    @check_closed
    def cursor(self):
        """Return a new Cursor Object using the connection."""
        cursor = Cursor(self)
        self.cursors.append(cursor)

        return cursor

    @check_closed
    def execute(self, operation, parameters=None):
        cursor = self.cursor()
        return cursor.execute(operation, parameters)

    def __enter__(self):
        return self.cursor()

    def __exit__(self, *exc):
        self.close()


class Cursor(object):
    """Connection cursor."""

    def __init__(self, connection):
        self.url = connection.url
        self.connection = connection

        # This read/write attribute specifies the number of rows to fetch at a
        # time with .fetchmany(). It defaults to 1 meaning to fetch a single
        # row at a time.
        self.arraysize = 1

        self.closed = False

        # this is updated only after a query
        self.description = None

        # this is set to an iterator after a successfull query
        self._results = None

    @property
    @check_result
    @check_closed
    def rowcount(self):
        # consume the iterator
        results = list(self._results)
        n = len(results)
        self._results = iter(results)
        return n

    @check_closed
    def close(self):
        """Close the cursor."""
        self.closed = True

    def is_supported_query(self, parsed):
        if hasattr(parsed, "tokens"):
            for token in parsed.tokens:
                if "on" == token.value.lower():
                    return True
        return False

    def get_supported_query(self, parsed, original_parsed=None):
        if not original_parsed:
            original_parsed = parsed
        # query is not supported. Try finding a supported inner query
        if parsed and hasattr(parsed, "tokens"):
            for token in parsed.tokens:
                value = token.value
                if value.replace("(", "").replace(" ", "").lower().startswith("select"):
                    if self.is_supported_query(token):
                        # found supported query
                        # return it and modified
                        return (value.strip(' ()') if value.strip(' ').startswith("(") else value,
                                original_parsed.value.replace(value, "(SELECT * FROM Model)"))
                    else:
                        ret = self.get_supported_query(token, original_parsed)
                        if ret:
                            return ret
                else:
                    ret = self.get_supported_query(token, original_parsed)
                    if ret:
                        return ret
        return None

    def execute_one_olap(self, operation, schema):
        # `_stream_query` returns a generator that produces the rows; we need
        # to consume the first row so that `description` is properly set, so
        # let's consume it and insert it back.
        results = self._stream_query(operation, schema)
        first_row = next(results)
        self._results = itertools.chain([first_row], results)

        return self._results

    @check_closed
    def execute(self, operation, parameters=None, schema=None, **kwargs):
        operation = apply_parameters(operation, parameters or {})
        # `_stream_query` returns a generator that produces the rows; we need
        # to consume the first row so that `description` is properly set, so
        # let's consume it and insert it back.
        from_sqlite = self._stream_query_sqlite(operation, schema)
        if not from_sqlite:
            results = self.execute_one_olap(operation, schema)
        else:
            results = self.from_sqlite_engine()

        first_row = next(results)
        self._results = itertools.chain([first_row], results)

        return self

    def from_sqlite_engine(self):
        with self.sqliteengine.connect() as connection:
            results = connection.execute(self.query_to_execute_on_db)
            self.description = results.cursor.description
            for row in results:
                yield row

    @check_closed
    def executemany(self, operation, seq_of_parameters=None):
        raise NotSupportedError(
            '`executemany` is not supported, use `execute` instead')

    @check_result
    @check_closed
    def fetchone(self):
        """
        Fetch the next row of a query result set, returning a single sequence,
        or `None` when no more data is available.
        """
        try:
            return self.next()
        except StopIteration:
            return None

    @check_result
    @check_closed
    def fetchmany(self, size=None):
        """
        Fetch the next set of rows of a query result, returning a sequence of
        sequences (e.g. a list of tuples). An empty sequence is returned when
        no more rows are available.
        """
        size = size or self.arraysize
        return list(itertools.islice(self, size))

    @check_result
    @check_closed
    def fetchall(self):
        """
        Fetch all (remaining) rows of a query result, returning them as a
        sequence of sequences (e.g. a list of tuples). Note that the cursor's
        arraysize attribute can affect the performance of this operation.
        """
        return list(self)

    @check_closed
    def setinputsizes(self, sizes):
        # not supported
        pass

    @check_closed
    def setoutputsizes(self, sizes):
        # not supported
        pass

    @check_closed
    def __iter__(self):
        return self

    @check_closed
    def __next__(self):
        return next(self._results)

    next = __next__

    def _stream_query(self, query, schema):
        """
        Stream rows from a query.

        This method will yield rows as the data is returned in chunks from the
        server.
        """
        self.description = None
        if query:
            res = self.connection.xmla.Execute(query, Catalog=schema)
            if hasattr(res, "getSlice"):
                xmla_slice = res.getSlice()

                # raise any error messages
                # if res.status_code != 200:
                #     payload = res.json()
                #     msg = (
                #         f'{payload["error"]} ({payload["errorClass"]}): ' +
                #         f'{payload["errorMessage"]}'
                #     )
                #     raise ProgrammingError(msg)

                Row = None
                axis_tuple = res.getAxisTuple(0)
                axis_tuple_1 = res.getAxisTuple(1)
                len_xmla_slice = len(xmla_slice)
                if len_xmla_slice > 0:
                    for irow in range(len_xmla_slice):
                        row = xmla_slice[irow]
                        if len(row) and isinstance(row[0], list):
                            row = row[0]
                        if len(axis_tuple_1) > irow:
                            first_col = axis_tuple_1[irow].get('Caption')
                        else:
                            first_col = None
                        # update description
                        if self.description is None:
                            self.description = get_description_from_row(row, res)

                        # return row in namedtuple
                        if Row is None:
                            keys = [d[0] for d in self.description]

                            Row = namedtuple('Row', keys, rename=True)
                        values = []
                        values.append(first_col)

                        for i in range(min(len(row), MAX_COLS - 1)):
                            if len(axis_tuple) > i:
                                c = row[i]
                                t = c.get('FmtValue')
                                values.append(t)
                        yield Row(*values)
            else:
                Row = None
                for irow in range(len(res.rows)):
                    row = res.rows[irow]
                    # update description
                    if self.description is None:
                        self.description = get_description_from_rowset(res.description)

                    # return row in namedtuple
                    if Row is None:
                        keys = [d[0] for d in self.description]
                        Row = namedtuple('Row', keys, rename=True)

                    yield Row(*row)
        else:
            Row = namedtuple('Row', ['All'], rename=True)

            yield Row(*['1'])

    def _stream_query_sqlite(self, query, schema):
        """
        Stream rows from a query.

        This method will yield rows as the data is returned in chunks from the
        server.
        """
        if query:
            statements = sqlparse.split(query)
            if len(statements) > 1:
                logger.warning("Multiple queries not supported")
            statement = statements[0]
            parsed = sqlparse.parse(statement)[0]
            # check if we have a containing unsupported select
            if not self.is_supported_query(parsed):
                queries = self.get_supported_query(parsed)
                if queries:
                    self.query_to_execute_on_olap = queries[0]
                    self.query_to_execute_on_db = queries[1]
                    results = self.execute_one_olap(self.query_to_execute_on_olap, schema)
                    self.sqliteengine = create_engine('sqlite:///:memory:', echo=True)
                    with self.sqliteengine.connect() as connection:
                        connection.execute("drop table if exists model")
                    df = pd.DataFrame(results)
                    df.to_sql('Model', self.sqliteengine.engine)
                    return True
        return False


def apply_parameters(operation, parameters):
    escaped_parameters = {
        key: escape(value) for key, value in parameters.items()
    }
    return operation % escaped_parameters


def escape(value):
    if value == '*':
        return value
    elif isinstance(value, string_types):
        return "'{}'".format(value.replace("'", "''"))
    elif isinstance(value, (int, float)):
        return value
    elif isinstance(value, bool):
        return 'TRUE' if value else 'FALSE'
    elif isinstance(value, (list, tuple)):
        return ', '.join(escape(element) for element in value)



from __future__ import unicode_literals

import os
import sys

from prompt_toolkit import prompt, AbortAction
from prompt_toolkit.history import FileHistory
from prompt_toolkit.contrib.completers import WordCompleter
from pygments.lexers import SqlLexer
from pygments.style import Style
from pygments.token import Token
from pygments.styles.default import DefaultStyle
from six.moves.urllib import parse
from tabulate import tabulate

from superset.db_engines.ssas.db import connect


keywords = [
    'EXPLAIN PLAN FOR',
    'WITH',
    'SELECT',
    'ALL',
    'DISTINCT',
    'FROM',
    'WHERE',
    'GROUP BY',
    'HAVING',
    'ORDER BY',
    'ASC',
    'DESC',
    'LIMIT',
]

aggregate_functions = [
    'COUNT',
    'SUM',
    'MIN',
    'MAX',
    'AVG',
    'APPROX_COUNT_DISTINCT',
    'APPROX_QUANTILE',
]

numeric_functions = [
    'ABS',
    'CEIL',
    'EXP',
    'FLOOR',
    'LN',
    'LOG10',
    'POW',
    'SQRT',
]

string_functions = [
    'CHARACTER_LENGTH',
    'LOOKUP',
    'LOWER',
    'REGEXP_EXTRACT',
    'REPLACE',
    'SUBSTRING',
    'TRIM',
    'BTRIM',
    'RTRIM',
    'LTRIM',
    'UPPER',
]

time_functions = [
    'CURRENT_TIMESTAMP',
    'CURRENT_DATE',
    'TIME_FLOOR',
    'TIME_SHIFT',
    'TIME_EXTRACT',
    'TIME_PARSE',
    'TIME_FORMAT',
    'MILLIS_TO_TIMESTAMP',
    'TIMESTAMP_TO_MILLIS',
    'EXTRACT',
    'FLOOR',
    'CEIL',
]

other_functions = [
    'CAST',
    'CASE',
    'WHEN',
    'THEN',
    'END',
    'NULLIF',
    'COALESCE',
]


class DocumentStyle(Style):
    styles = {
        Token.Menu.Completions.Completion.Current: 'bg:#00aaaa #000000',
        Token.Menu.Completions.Completion: 'bg:#008888 #ffffff',
        Token.Menu.Completions.ProgressButton: 'bg:#003333',
        Token.Menu.Completions.ProgressBar: 'bg:#00aaaa',
    }
    styles.update(DefaultStyle.styles)


def get_connection_kwargs(url):
    parts = parse.urlparse(url)
    if ':' in parts.netloc:
        host, port = parts.netloc.split(':', 1)
        port = int(port)
    else:
        host = parts.netloc
        port = 8082

    return {
        'host': host,
        'port': port,
        'path': parts.path,
        'scheme': parts.scheme,
    }


def get_tables(connection):
    cursor = connection.cursor()
    return [
        row.TABLE_NAME for row in
        cursor.execute('SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES')
    ]


def get_autocomplete(connection):
    return (
        keywords +
        aggregate_functions +
        numeric_functions +
        string_functions +
        time_functions +
        other_functions +
        get_tables(connection)
    )


def main():
    history = FileHistory(os.path.expanduser('~/.netas_olap_dbapi_history'))

    try:
        url = sys.argv[1]
    except IndexError:
        url = 'http://localhost:80/OLAP/msmdpump.dll'
    kwargs = get_connection_kwargs(url)
    connection = connect(**kwargs)
    cursor = connection.cursor()

    words = get_autocomplete(connection)
    sql_completer = WordCompleter(words, ignore_case=True)

    while True:
        try:
            query = prompt(
                '> ', lexer=SqlLexer, completer=sql_completer,
                style=DocumentStyle, history=history,
                on_abort=AbortAction.RETRY)
        except EOFError:
            break  # Control-D pressed.

        # run query
        if query.strip():
            try:
                result = cursor.execute(query.rstrip(';'))
            except Exception as e:
                print(e)
                continue

            headers = [t[0] for t in cursor.description or []]
            print(tabulate(result, headers=headers))

    print('GoodBye!')


if __name__ == '__main__':
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from sqlalchemy.engine import default
from sqlalchemy.sql import compiler
from sqlalchemy import types,util

import superset.db_engines.ssas as db
from superset.db_engines.ssas import exceptions


RESERVED_SCHEMAS = ['INFORMATION_SCHEMA']
"""
https://docs.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas/aee32107-fde6-43bd-beeb-46e82851abf4

0 – (DBTYPE_EMPTY) Indicates that no value was specified.
2 – (DBTYPE_I2) Indicates a two-byte signed integer.
3 – (DBTYPE_I4) Indicates a four-byte signed integer.
4 – (DBTYPE_R4) Indicates a single-precision floating-point value.
5 – (DBTYPE_R8) Indicates a double-precision floating-point value.
6 – (DBTYPE_CY) Indicates a currency value. Currency is a fixed-point number with four digits to the right of the decimal point and is stored in an eight-byte signed integer scaled by 10,000.
7 – (DBTYPE_DATE) Indicates a date value. Date values are stored as Double, the whole part of which is the number of days since December 30, 1899, and the fractional part of which is the fraction of a day.
8 – (DBTYPE_BSTR) A pointer to a BSTR, which is a null-terminated character string in which the string length is stored with the string.
9 – (DBTYPE_IDISPATCH) Indicates a pointer to an IDispatch interface on an OLE object.
10 – (DBTYPE_ERROR) Indicates a 32-bit error code.
11 – (DBTYPE_BOOL) Indicates a Boolean value.
12 – (DBTYPE_VARIANT) Indicates an Automation variant.
13 – (DBTYPE_IUNKNOWN) Indicates a pointer to an IUnknown interface on an OLE object.
14 – (DBTYPE_DECIMAL) Indicates an exact numeric value with a fixed precision and scale. The scale is between 0 and 28.
16 – (DBTYPE_I1) Indicates a one-byte signed integer.
17 – (DBTYPE_UI1) Indicates a one-byte unsigned integer.
18 – (DBTYPE_UI2) Indicates a two-byte unsigned integer.
19 – (DBTYPE_UI4) Indicates a four-byte unsigned integer.
20 – (DBTYPE_I8) Indicates an eight-byte signed integer.
21 – (DBTYPE_UI8) Indicates an eight-byte unsigned integer.
72 – (DBTYPE_GUID) Indicates a GUID.
128 – (DBTYPE_BYTES) Indicates a binary value.
129 – (DBTYPE_STR) Indicates a string value.
130 – (DBTYPE_WSTR) Indicates a null-terminated Unicode character string.
131 – (DBTYPE_NUMERIC) Indicates an exact numeric value with a fixed precision and scale. The scale is between 0 and 38.
132 – (DBTYPE_UDT) Indicates a user-defined variable.
133 – (DBTYPE_DBDATE) Indicates a date value (yyyymmdd).
134 – (DBTYPE_DBTIME) Indicates a time value (hhmmss).
135 – (DBTYPE_DBTIMESTAMP) Indicates a date-time stamp (yyyymmddhhmmss plus a fraction in billionths).
136 - (DBTYPE_HCHAPTER) Indicates a four-byte chapter value used to identify rows in a child rowset.
"""

type_map = {
"0" : types.String,  #(DBTYPE_EMPTY) Indicates that no value was specified.
"2" : types.BigInteger,  #(DBTYPE_I2) Indicates a two-byte signed integer.
"3" : types.BigInteger,  #(DBTYPE_I4) Indicates a four-byte signed integer.
"4" : types.Float,  #(DBTYPE_R4) Indicates a single-precision floating-point value.
"5" : types.Float,  #(DBTYPE_R8) Indicates a double-precision floating-point value.
"6" : types.Float,  #(DBTYPE_CY) Indicates a currency value. Currency is a fixed-point number with four digits to the right of the decimal point and is stored in an eight-byte signed integer scaled by 10,000.
"7" : types.Float,  #(DBTYPE_DATE) Indicates a date value. Date values are stored as Double, the whole part of which is the number of days since December 30, 1899, and the fractional part of which is the fraction of a day.
"8" : types.String,  #(DBTYPE_BSTR) A pointer to a BSTR, which is a null-terminated character string in which the string length is stored with the string.
"9" : types.BigInteger,  #(DBTYPE_IDISPATCH) Indicates a pointer to an IDispatch interface on an OLE object.
"10" : types.BigInteger,  #(DBTYPE_ERROR) Indicates a 32-bit error code.
"11" : types.Boolean,  #(DBTYPE_BOOL) Indicates a Boolean value.
"12" : types.String,  #(DBTYPE_VARIANT) Indicates an Automation variant.
"13" : types.String,  #(DBTYPE_IUNKNOWN) Indicates a pointer to an IUnknown interface on an OLE object.
"14" : types.Float,  #(DBTYPE_DECIMAL) Indicates an exact numeric value with a fixed precision and scale. The scale is between 0 and 28.
"16" : types.BigInteger,  #(DBTYPE_I1) Indicates a one-byte signed integer.
"17" : types.BigInteger,  #(DBTYPE_UI1) Indicates a one-byte unsigned integer.
"18" : types.BigInteger,  #(DBTYPE_UI2) Indicates a two-byte unsigned integer.
"19" : types.BigInteger,  #(DBTYPE_UI4) Indicates a four-byte unsigned integer.
"20" : types.BigInteger,  #(DBTYPE_I8) Indicates an eight-byte signed integer.
"21" : types.BigInteger,  #(DBTYPE_UI8) Indicates an eight-byte unsigned integer.
"72" : types.String,  #(DBTYPE_GUID) Indicates a GUID.
"128" : types.String,  #(DBTYPE_BYTES) Indicates a binary value.
"129" : types.String,  #(DBTYPE_STR) Indicates a string value.
"130" : types.String,  #(DBTYPE_WSTR) Indicates a null-terminated Unicode character string.
"131" : types.Float,  #(DBTYPE_NUMERIC) Indicates an exact numeric value with a fixed precision and scale. The scale is between 0 and 38.
"132" : types.String,  #(DBTYPE_UDT) Indicates a user-defined variable.
"133" : types.String,  #(DBTYPE_DBDATE) Indicates a date value (yyyymmdd).
"134" : types.String,  #(DBTYPE_DBTIME) Indicates a time value (hhmmss).
"135" : types.String,  #(DBTYPE_DBTIMESTAMP) Indicates a date-time stamp (yyyymmddhhmmss plus a fraction in billionths).
"136" : types.String,  # (DBTYPE_HCHAPTER) Indicates a four-byte chapter value used to identify rows in a child rowset.
 
}
# type_map = {
#     'char': types.String,
#     'varchar': types.String,
#     'float': types.Float,
#     'decimal': types.Float,
#     'real': types.Float,
#     'double': types.Float,
#     'boolean': types.Boolean,
#     'tinyint': types.BigInteger,
#     'smallint': types.BigInteger,
#     'integer': types.BigInteger,
#     'bigint': types.BigInteger,
#     'timestamp': types.TIMESTAMP,
#     'date': types.DATE,
# }


class UniversalSet(object):
    def __contains__(self, item):
        return True


class SSASIdentifierPreparer(compiler.IdentifierPreparer):
    reserved_words = UniversalSet()


class SSASCompiler(compiler.SQLCompiler):
    pass


class SSASTypeCompiler(compiler.GenericTypeCompiler):
    def visit_REAL(self, type_, **kwargs):
        return "DOUBLE"

    def visit_NUMERIC(self, type_, **kwargs):
        return "LONG"

    visit_DECIMAL = visit_NUMERIC
    visit_INTEGER = visit_NUMERIC
    visit_SMALLINT = visit_NUMERIC
    visit_BIGINT = visit_NUMERIC
    visit_BOOLEAN = visit_NUMERIC
    visit_TIMESTAMP = visit_NUMERIC
    visit_DATE = visit_NUMERIC

    def visit_CHAR(self, type_, **kwargs):
        return "STRING"

    visit_NCHAR = visit_CHAR
    visit_VARCHAR = visit_CHAR
    visit_NVARCHAR = visit_CHAR
    visit_TEXT = visit_CHAR

    def visit_DATETIME(self, type_, **kwargs):
        raise exceptions.NotSupportedError('Type DATETIME is not supported')

    def visit_TIME(self, type_, **kwargs):
        raise exceptions.NotSupportedError('Type TIME is not supported')

    def visit_BINARY(self, type_, **kwargs):
        raise exceptions.NotSupportedError('Type BINARY is not supported')

    def visit_VARBINARY(self, type_, **kwargs):
        raise exceptions.NotSupportedError('Type VARBINARY is not supported')

    def visit_BLOB(self, type_, **kwargs):
        raise exceptions.NotSupportedError('Type BLOB is not supported')

    def visit_CLOB(self, type_, **kwargs):
        raise exceptions.NotSupportedError('Type CBLOB is not supported')

    def visit_NCLOB(self, type_, **kwargs):
        raise exceptions.NotSupportedError('Type NCBLOB is not supported')


class SSASDialect(default.DefaultDialect):

    name = 'sass'
    scheme = 'http'
    driver = 'rest'
    preparer = SSASIdentifierPreparer
    statement_compiler = SSASCompiler
    type_compiler = SSASTypeCompiler
    supports_alter = False
    supports_pk_autoincrement = False
    supports_default_values = False
    supports_empty_insert = False
    supports_unicode_statements = True
    supports_unicode_binds = True
    returns_unicode_strings = True
    description_encoding = None
    supports_native_boolean = True
    _has_events = True
    
    @classmethod
    def dbapi(cls):
        return db

    def create_connect_args(self, url):
        kwargs = {
            'host': url.host,
            'port': url.port or 80,
            'path': url.database,
            'scheme': self.scheme,
            'username': url.username,
            'password': url.password,
            'trusted_connection': url.query.get("trusted_connection") == "yes"
        }
        return ([], kwargs)

    def get_schema_names(self, connection, **kwargs):
        # Each SSAS datasource appears as a table in the "SSAS" schema. This
        # is also the default schema, so SSAS datasources can be referenced as
        # either SSAS.dataSourceName or simply dataSourceName.
        result = connection.raw_connection().connection.xmla.getDBSchemaCatalogs()

        return [
            row.CATALOG_NAME for row in result
            if row.CATALOG_NAME not in RESERVED_SCHEMAS
        ]

    def has_table(self, connection, table_name, schema=None):
        if schema:
            result = connection.raw_connection().connection.xmla.getMDSchemaDimensions(properties={"Catalog":schema})
        else:
            result = connection.raw_connection().connection.xmla.getMDSchemaDimensions()
        return len([
            row.DIMENSION_NAME for row in result if row.DIMENSION_NAME == table_name
        ])>0
         

    def get_table_names(self, connection, schema=None, **kwargs):
        if schema:
            result = connection.raw_connection().connection.xmla.getMDSchemaDimensions(properties={"Catalog":schema})
        else:
            result = connection.raw_connection().connection.xmla.getMDSchemaDimensions()
        return [  row.DIMENSION_NAME for row in result   ]

    def get_view_names(self, connection, schema=None, **kwargs):
        return []

    def get_table_options(self, connection, table_name, schema=None, **kwargs):
        return {}

    def get_columns(self, connection, table_name, schema=None, **kwargs):
         
        if schema:
            result = connection.raw_connection().connection.xmla.getMDSchemaLevels(properties={f"Catalog":f"{schema}"})
        else:
            result = connection.raw_connection().connection.xmla.getMDSchemaLevels(properties={f"Catalog":f"{schema}"})
        if table_name:
            result =  [row for row in result if row.get("DIMENSION_UNIQUE_NAME").strip("[]") == table_name] 

        return [
            {
                'name': row.get("LEVEL_UNIQUE_NAME"),
                'type': type_map[row.get("LEVEL_DBTYPE")],
                'nullable': False,
                'default': None #get_default(row.COLUMN_DEFAULT),
            }
            for row in result
        ]

    def get_pk_constraint(self, connection, table_name, schema=None, **kwargs):
        return {'constrained_columns': [], 'name': None}

    def get_foreign_keys(self, connection, table_name, schema=None, **kwargs):
        return []

    def get_check_constraints(
        self,
        connection,
        table_name,
        schema=None,
        **kwargs
    ):
        return []

    def get_table_comment(self, connection, table_name, schema=None, **kwargs):
        return {'text': ''}

    def get_indexes(self, connection, table_name, schema=None, **kwargs):
        return []

    def get_unique_constraints(
        self,
        connection,
        table_name,
        schema=None,
        **kwargs
    ):
        return []

    def get_view_definition(
        self,
        connection,
        view_name,
        schema=None,
        **kwargs
    ):
        pass

    def do_rollback(self, dbapi_connection):
        pass

    def _check_unicode_returns(self, connection, additional_tests=None):
        return True

    def _check_unicode_description(self, connection):
        return True

    @util.memoized_property
    def _dialect_specific_select_one(self):
        return ""

SSASHTTPDialect = SSASDialect


class SSASHTTPSDialect(SSASDialect):

    scheme = 'https'


def get_is_nullable(SSAS_is_nullable):
    # this should be 'YES' or 'NO'; we default to no
    return SSAS_is_nullable.lower() == 'yes'


def get_default(SSAS_column_default):
    # currently unused, returns ''
    return str(SSAS_column_default) if SSAS_column_default != '' else None
